IMPLEMENTATION MODULE MiscPMOS;

        (************************************************)
        (*                                              *)
        (*      Miscellaneous PMOS procedures           *)
        (*                                              *)
        (*  Programmer:         P. Moylan               *)
        (*  Last edited:        23 October 1996         *)
        (*  Status:             Just started XDS port   *)
        (*                                              *)
        (************************************************)

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, Obtain, Release;

VAR mutex: Lock;

(************************************************************************)
(*                              STRING COPY                             *)
(************************************************************************)

PROCEDURE CopyString (source: ARRAY OF CHAR;  VAR (*OUT*) dest: ARRAY OF CHAR);

    (* Copies a string, with truncation or null termination as needed.  *)
    (* This function is provided in order to help software portability, *)
    (* i.e. to avoid having to rewrite code for no reason other than    *)
    (* a change of compilers.                                           *)

    VAR j, last: CARDINAL;  AddNull: BOOLEAN;

    BEGIN
        last := HIGH(dest);
        AddNull := HIGH(source) < last;
        IF AddNull THEN last := HIGH(source) END (*IF*);
        FOR j := 0 TO last DO dest[j] := source[j] END (*FOR*);
        IF AddNull THEN dest[last+1] := CHR(0) END (*IF*);
    END CopyString;

(************************************************************************)
(*                   MISCELLANEOUS LOW-LEVEL OPERATIONS                 *)
(************************************************************************)

PROCEDURE EnterCriticalSection (): CARDINAL;

    (* Saves the processor flags word, including the current "interrupt *)
    (* enable" status, on the caller's stack, and returns with          *)
    (* interrupts disabled.   NOTE: this procedure and the following    *)
    (* one should be used as a matched pair.                            *)

    BEGIN
        Obtain (mutex);
        RETURN 0;
    END EnterCriticalSection;

(************************************************************************)

PROCEDURE LeaveCriticalSection (SavedProcessorStatus: CARDINAL);

    (* Restores the processor flags word, including the "interrupt      *)
    (* enable" status, from the stack.  NOTE: this procedure and the    *)
    (* one above should be used as a matched pair.                      *)

    BEGIN
        Release (mutex);
    END LeaveCriticalSection;

(************************************************************************)

BEGIN
    CreateLock (mutex);
END MiscPMOS.
