IMPLEMENTATION MODULE Semaphores;

        (********************************************************)
        (*                                                      *)
        (*      Implements the Wait and Signal operations on    *)
        (*      semaphores.                                     *)
        (*                                                      *)
        (*      Programmer:     P. Moylan                       *)
        (*      Last edited:    18 September 1996               *)
        (*      Status:         Problems!  Needs re-design      *)
        (*                                                      *)
        (*          This is a quick-and-dirty version, and      *)
        (*          still has room for improvement.             *)
        (*                                                      *)
        (*          Now under revision, as I've discovered      *)
        (*          that OS/2 event semaphores don't work       *)
        (*          like ordinary semaphores.                   *)
        (*                                                      *)
        (*          After latest change, TaskTest exits         *)
        (*          without even running the child threads,     *)
        (*          so something is fishy here.                 *)
        (*          Analysis: the problem here is that an OS/2  *)
        (*          mutex semaphore is subtly different from    *)
        (*          a traditional binary semaphore.  I'm going  *)
        (*          to have to look for a different approach,   *)
        (*          possibly using message queues.              *)
        (*                                                      *)
        (*          Current idea: queues will probably not do   *)
        (*          the job satisfactorily;  but I think I can  *)
        (*          find a solution using DosSuspendThread.     *)
        (*                                                      *)
        (********************************************************)

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM TerminationControl IMPORT
    (* proc *)  Crash;

FROM OS2 IMPORT
    (* const*)  SEM_INDEFINITE_WAIT,
    (* type *)  HMTX, HEV,
    (* proc *)  DosCreateMutexSem, DosCloseMutexSem, DosCreateEventSem,
                DosRequestMutexSem, DosReleaseMutexSem,
                DosWaitEventSem, DosPostEventSem;

(************************************************************************)

TYPE Semaphore = POINTER TO
                    RECORD
                        value: INTEGER;
                        access: HMTX;
                        EventSem: (*HEV*)HMTX;
                    END;

(************************************************************************)

PROCEDURE CreateSemaphore (VAR (*OUT*) s: Semaphore;
                                        InitialValue: CARDINAL);

    (* Creates semaphore s, with the given initial value and an empty   *)
    (* queue.                                                           *)

    VAR OK: BOOLEAN;

    BEGIN
        NEW(s);
        WITH s^ DO
            OK := (DosCreateMutexSem (NIL, access, 0, TRUE) = 0)
                       AND (DosCreateMutexSem(*DosCreateEventSem*) (NIL, EventSem, 0, TRUE) = 0);
            value := InitialValue;
            DosReleaseMutexSem (access);
        END (*WITH*);
        IF NOT OK THEN
            Crash ("Couldn't create semaphore.");
        END (*IF*);
    END CreateSemaphore;

(************************************************************************)

PROCEDURE DestroySemaphore (VAR (*INOUT*) s: Semaphore);

    (* Reclaims any space used by semaphore s.  Remark:  It is not at   *)
    (* all obvious what should be done with any tasks which happen to   *)
    (* be blocked on this semaphore (should they be unblocked, or       *)
    (* killed?).  At present we take the easy way out and assume that   *)
    (* there are no pending operations on s at the time that it is      *)
    (* destroyed.                                                       *)

    BEGIN
        WITH s^ DO
            DosCloseMutexSem (access);
        END (*WITH*);
        DISPOSE (s);
    END DestroySemaphore;

(************************************************************************)

PROCEDURE Wait (s: Semaphore);

    (* Decrements the semaphore value.  If the value goes negative, the *)
    (* calling task is blocked and there is a task switch.              *)

    BEGIN
        IF s = NIL THEN Crash ("Wait on nonexistent semaphore"); END(*IF*);
        WITH s^ DO
            DosRequestMutexSem (access, SEM_INDEFINITE_WAIT);
            DEC (value);
            IF value < 0 THEN
                DosReleaseMutexSem (access);
                (*DosWaitEventSem (EventSem, SEM_INDEFINITE_WAIT);*)
                DosRequestMutexSem (EventSem, SEM_INDEFINITE_WAIT);
            ELSE
                DosReleaseMutexSem (access);
            END (*IF*);
        END (*WITH*);
    END Wait;

(************************************************************************)

PROCEDURE TimedWaitT (s: Semaphore;  TimeLimit: INTEGER;
                        VAR (*OUT*) TimedOut: BOOLEAN);

    (* Like procedure Wait, except that it returns with TimedOut TRUE   *)
    (* if the corresponding Signal does not occur within TimeLimit      *)
    (* clock ticks.                                                     *)

    BEGIN
        IF s = NIL THEN Crash ("Wait on nonexistent semaphore"); END(*IF*);
        WITH s^ DO
            DosRequestMutexSem (access, SEM_INDEFINITE_WAIT);
            DEC (value);
            IF value < 0 THEN
                DosReleaseMutexSem (access);
                (*DosWaitEventSem (EventSem, TimeLimit);*)
                DosRequestMutexSem (EventSem, TimeLimit);
            ELSE
                DosReleaseMutexSem (access);
            END (*IF*);
        END (*WITH*);
    END TimedWaitT;

(************************************************************************)

PROCEDURE Signal (s: Semaphore);

    (* Increments the semaphore value.  Unblocks one task, if there was *)
    (* one waiting on this semaphore.                                   *)

    BEGIN
        IF s = NIL THEN Crash ("Signal on nonexistent semaphore"); END(*IF*);
        WITH s^ DO
            DosRequestMutexSem (access, SEM_INDEFINITE_WAIT);
            INC (value);
            IF value <= 0 THEN
                DosReleaseMutexSem (access);
                (*DosPostEventSem (EventSem);*)
                DosReleaseMutexSem (EventSem);
            ELSE
                DosReleaseMutexSem (access);
            END (*IF*);
        END (*WITH*);
    END Signal;

END Semaphores.

